import { ReflectedAttrToProps } from '../../_global/types';
/**
 * This module create property instance and get/set for all attributes and observedAttributes
 */
export declare abstract class BootstrapClass extends HTMLElement {
    [key: string]: any;
    /**
     * This is re-usable static method, which sets property on the instance prototype based on the given attribute name
     * @param {string} name attribute name
     */
    defineProp(name: string, reflectedAttrToProps: any): void;
    /**
     * This loops through all the given attributes, observedAttributes and send it to `defineProp()`
     * @param {object} attributes list of attributes added to the component
     * @param {array} observedAttributes list of observed attributes
     */
    createAttributesToProperties(reflectedAttrToProps: ReflectedAttrToProps): void;
    /**
     * has attribute
     * @param {string} attribute name
     * @param {object} target DOM element
     */
    hasAttr(attr: string, target?: HTMLElement): boolean;
    /**
     * set attribute value
     * @param {string} attribute name
     * @param {object} target DOM element
     */
    getAttr(attr: string, target?: HTMLElement): string | null;
    /**
     * set attribute value
     * @param {string} attribute name
     * @param {string} new/changed value for the attribute
     * @param {object} target DOM element
     */
    setAttr(attr: string, val: string, target?: HTMLElement): void;
    /**
     * remove attribute
     * @param {string} attribute name
     * @param {object} target DOM element
     */
    removeAttr(attr: string, target?: HTMLElement): void;
    /**
     * helper function to add class name using classList
     * @param {string} className string of class name
     * @param {object} target DOM element
     * @return {object} element DOM element
     */
    addClass(className: string, target?: HTMLElement): HTMLElement;
    /**
     * helper function to remove class name using classList
     * @param {string} className string of class name
     * @param {object} target DOM element
     * @return {object} element DOM element
     */
    removeClass(className: string, target?: HTMLElement): HTMLElement;
    /**
     * helper function to check for has class name using classList
     * @param {string} className string of class name
     * @param {object} target DOM element
     * @return {boolean} contains true/false
     */
    hasClass(className: string, target?: HTMLElement): boolean;
    /**
     * helper function to toggle class name using classList
     * @param {string} className string of class name
     * @param {object} target DOM element
     * @param {boolean} force boolean to use force optionally
     * @return {object} element DOM element
     */
    toggleClass(className: string, target: HTMLElement | undefined, force: boolean): boolean;
    /**
     * helper to find child nodes rendered inside
     * @param {string} selector that is used to search the DOM node
     * @param {object} target Element in which search should be performed
     * @return {object} Element
     */
    find(selector: string, target?: HTMLElement): HTMLElement | null;
    /**
     * helper to find all child nodes rendered inside
     * @param {string} selector that is used to search the DOM node
     * @param {object} target Element in which search should be performed
     * @return {array} Elements that are matched
     */
    findAll(selector: string, target?: HTMLElement): NodeListOf<HTMLElement>;
    /**
     * simplyfied event handler function to add events
     * @param {string} eventName event type ex: click, blur, etc.
     * @param {object} target HTML element that needs event to be added
     * @param {function | Object} callback function or Object (which has handleEvent method) that needs to be triggerd
     */
    on(eventName: string, target: HTMLElement | undefined, callback: EventListenerOrEventListenerObject): void;
    /**
     * simplyfied event handler function to remove events
     * @param {string} eventName event type ex: click, blur, etc.
     * @param {object} target HTML element that needs event to be added
     * @param {function | Object} callback function or Object (which has handleEvent method) that needs to be triggerd
     */
    off(eventName: string, target: HTMLElement | undefined, callback: EventListenerOrEventListenerObject): void;
    /**
     * trigger native/custom event and pass data between components
     * @param {string} eventName custom event name
     * @param {object} target element reference on which event needs to be triggerd
     * @param {object} eventData custom event data to share
     */
    trigger(eventName: string, target?: HTMLElement, eventData?: {}): void;
}
export default BootstrapClass;
